"""
Built-in configuration libary for use with server-client model instruments.

Currently supports toml syntax but internally data is treated as nested 
dictionaries(tables) or lists of dicts (arrays of tables) so other syntax are
possible by writing a parser.

decoder -  Decoder object responsible for loading data into nested dicts structure
encoder -  Encoder object for writing nested dicts to string/file.
types   -  Templating objects used to add comments and define fomrating to use.

config - Config object provides a top-level interface, combining data and 
         template, together with pretty printing and automatic validation of 
         values against template.
"""

#import the parser
from . import parser
from .parser import load, save

# import the Config wrapper class for data and templating classes
from . import types
from .config import Config



"""
TOML decoder using toml format https://toml.io
-----------------------------------------------

TomlDecoder class - main decoder/parsing class use open/load/loads to load 
                    configuration from filepath, file or string.
                    
Templating: 
-----------
The decoder also generates a template containing formating and
comments found in the file. This can be passed to the encoder to retain comments
and some formatting styles (e.g. int written as hexadecimal).

This template is a nested dictionary/TableType of TomlType objects defined for 
the types in the TOML syntax.

Comments are treated as being associated with a value using the following rules.
 - A full comments line (or multiple full comment lines) are associated with the 
    next key/value or table.
 - Inline comments are associated with the key/value on the same line.
 - Comments inside arrays (lists) or inline tables (dicts) are associated with  
    the preceeding value),
"""
import io                                             #for string buffers
import re                                             #for spliting of values
from dateutil.parser import parse as datetime_parser  #for easy datetime parsing
from ast import literal_eval         #processing of escape sequences in strings

from .types import *                 #types for template creation on load

__all__ = ['toml_parser', 'toml_writer', 'load', 'save']

#%%-----------------------------------------------------------------------------
# Toml parser class - using 
#-------------------------------------------------------------------------------
class TomlParser():
    """
    Parser using a toml compatible syntax:

    Should support almost full toml spec now however there may be some issues.
    
    + This allows table names to have spaces (not in toml spec) don't use them
    if you want to be 100% compatible
    + Inline tables can have newlines inside {} (not in toml spec) but really 
    normal tables should be used.
    """
    def open(self, filename):
        """ Open the filename give and parses """
        with open(filename, 'r', newline='', encoding='utf-8') as f:
            return self.parse(f)
            
    def load(self, f):
        """
        Parse the file object
        """ 
        #wrap the file object buffer to ensure correct line-ending and encoding mode
        self.buffer = io.TextIOWrapper( f.buffer, newline='', encoding='utf-8')
        self.parse(f)
        
    def loads(self, s):
        """
        Parse the string
        """
        f = io.StringIO(s)
        return self.parse(f)

    # Main parsing method
    def parse(self, f):
        """
        Parse the file or buffer given return nested dicts
        
        data      = nested dictionary containing parsed data
        template  = nested dictionary containing comments and format information
        """ 
        # buffer some values internally for access
        self.data = {}   # data
        self.template = TableType()
                        # comments are associated with the next key or for
                        # inline comments the current key
        
        #wrap the buffer to ensure correct line endings
        self.buffer = f
        self.linenum = 0        #line count for error reporting

        #read top of filecomments (until a newline) or key/table is encountered
        while True:
            #read next line
            try:
                rawline = self._read_more()
            except EOFError:
                break
            
            line = rawline.strip()
            if line.startswith('#'):
                self.template.pre_comments.append(line)
            else:
                break
        #add a blank line comment after to seperate main table comments to key pre-comment
        self.template.pre_comments.append('')
        
        #current data and template tables
        self.curent_dtable = self.data
        self.curent_ttable = self.template
        self.pre_comments = []     #cached comments to be stored with next key
        
        #parse items until end_of_file
        # line from above is parsed first!
        while True:
            # ignore blank lines unless after comments
            if line=='' and len(self.pre_comments)==0:
                pass
            elif line=='' and len(self.pre_comments)>0:
                self.pre_comments.append('')

            #check for comment line
            elif line.startswith('#'):
                self.pre_comments.append(line)
                
            #check for array of tables (list of dicts)
            elif line.startswith('[['):
                self._parse_array_of_tables(line)
            
            #check for table
            elif line.startswith('['):
                self._parse_table(line)
            
            #check for key/value pairs
            else:
                
                #get key names list
                names, remainder = self._get_names( line, term='=')
                key = names[-1]
                tables = names[:-1]
                #print(f'Parse1: {names} {key} {tables}: {remainder}')
                
                #determine type
                value, ttype, remainder = self._parse_value(remainder) 
                #print(f'Parse: {key} value {value} : {repr(remainder)}')
                
                if len(tables)==0:
                    #store into current table
                    dtable = self.curent_dtable
                    ttable = self.curent_ttable
                else:
                    #get the correct table
                    dtable = self._get_data_table(tables, self.curent_dtable)
                    ttable = self._get_template_table(tables, self.curent_ttable)
                    
                # check for inline comment
                post_comment = self._parse_inline_comment(remainder)
                
                # store comments
                if ttype is None:
                    ttype = TomlType() # create tomltype object for key
                
                if self.pre_comments:
                    ttype.set_comments( self.pre_comments) # store comments
                    self.pre_comments = [] # and clear list
                    
                if post_comment:
                    ttype.add_comment(post_comment, post=True)
                    
                #check key is not already defined
                if key in dtable:
                    raise ParsingError(f'Keyname {key} already assigned!')
            
                #store value
                dtable[key] = value
                
                # store tomltype object into the template table
                ttable[key] = ttype

            #read next line
            try:
                rawline = self._read_more()
            except EOFError:
                break

            #strip whitespace
            line = rawline.strip()
            
        #clear cache and return
        data = self.data
        template = self.template
        self.data = None
        self.template = None
        self.curent_dtable = self.data
        self.curent_ttable = self.template
        self.pre_comments = [] 
        
        return data, template
    
    # general helpers
    def _read_more(self):
        """ Read the next line form the stream """
        line = self.buffer.readline()
        self.linenum +=1
        if line=='':
            raise EOFError()
        return line
    
    def _get_string(self, line):
        """ get a complete string from the line and return remainder """
        if line.startswith("'"):
            sep="'"
        elif line.startswith('"'):
            sep = '"'
        else:
            return None, line
        i=1
        escaped=False
        while i <len(line):
            if escaped==True:
                escaped=False
            elif line[i]=='\\':
                escaped=True
            elif line[i]==sep:
                break
            i+=1

        if i==len(line) and line[i]!=sep:
            raise ParsingError(f'Unclosed string {repr(sep)} at line number {self.linenum} ; {line}')
        value = line[1:i]
        remainder = line[i+1:]

        return value, remainder
    
    
    def _get_data_table(self, names, parent):
        """ Get or make the table structure given in names """
        # make correct nested dictionaries for this table
        dtable = parent
        for n in names:
            # get each parent table in turn or create
            if n not in dtable:
                dtable[n] = {}
                dtable = dtable[n]
                
            elif isinstance(dtable[n] , list):
                # get the last element of array of tables/dicts
                # eg { 'first' : [ { 'second' : {'returns':'this dict'},}, ]
                # 'first.second' 
                # 'first' returns { 'second' : {'returns':'this dict'},}
                # 'second' returns {'returns':'this dict'}
                dtable = dtable[n][-1]
                
            elif isinstance( dtable[n], dict):
                dtable = dtable[n]
            else:
                raise ParsingError(f'Table name "{n}" already has value at line number {self.linenum}; {line}')
        return dtable
        
    
    def _get_template_table(self, names, parent):
        """ Get or make the template table structure given in names """
        # make correct nested dictionaries for this table
        ttable = parent
        for n in names:
            # get each parent table in turn or create
            if n not in ttable:
                ttable[n] = TableType()
                ttable = ttable[n]
                
            elif isinstance(ttable[n] , ArrayType):
                # get the last element of array of tables/dicts
                # eg { 'first' : [ { 'second' : {'returns':'this dict'},}, ]
                # 'first.second' 
                # 'first' returns { 'second' : {'returns':'this dict'},}
                # 'second' returns {'returns':'this dict'}
                ttable = ttable[n][-1]
                
            elif isinstance( ttable[n], TableType):
                ttable = ttable[n]
            else:
                raise ParsingError(f'Table name "{n}" already has tomltype value at line number {self.linenum}; {line}')
        return ttable
    
    
    def _get_names(self, line, term=']'):
        """Get the list of names in a dotted table or key name"""
        #split out the full table name from inside [] or dotted key
    
        # can be quoted and dotted...
        # e.g. [table."[sub].".table]
        remainder = line # for table assumes first [ is stripped
        names = []
        while True:
            #print(names, remainder)
            
            #check for quoted part
            name,remainder = self._get_string(remainder)
            if name is not None:
                #store name
                names.append(name)
                
                #remove any whitespace and '.'
                remainder = remainder.lstrip()
                remainder = remainder.lstrip('.')
                
                #check for end ']]' ']' or '='
                if remainder.startswith(term):
                    remainder = remainder[len(term):] #any comment
                    break #leave search loop
            else:
                #try to split at . or term (']'|'=')
                idot = remainder.find('.')
                iterm = remainder.find(term)
                
                #neither found
                if (idot==-1) and (iterm==-1):
                    raise ParsingError(f'Missing {term} terminator, at line number {self.linenum}; {line}')

                #found term first
                elif (idot==-1) or (iterm<idot):
                    name = remainder[:iterm]
                    names.append( name.strip() )
                    remainder = remainder[iterm+len(term):]
                    break #leave search loop
                    
                #found . first
                elif (iterm==-1) or (idot<iterm):
                    name = remainder[:idot]
                    names.append( name.strip() )
                    remainder = remainder[idot+1:]
                    #remove any whitespace
                    remainder = remainder.lstrip()

            #check remainder for end conditions - this shouldn't happen?
            if remainder in ['', '\n']:
                raise ParsingError(f'Missing {term} terminator, at line number {self.linenum}; {line}')
        return names, remainder
    
    # Top level parsing functions
        
    def _parse_table(self, line):
        """
        Parse a sub table line entry
        """
        # Split out the full table name from inside []:
        # can be quoted and dotted...
        # e.g. [table."[sub].".table]
        #strip initial '[' 
        names, remainder = self._get_names(line[1:], term=']')
        #print(f'names {names}, remainder {repr(remainder)}')
        
        # Get or create new tables, and update the current tables
        self.curent_dtable = self._get_data_table(names, self.data)
        self.curent_ttable = self._get_template_table(names, self.template)
        
        #save any precomments
        if self.pre_comments:
            self.curent_ttable.set_comments( self.pre_comments)
            self.pre_comments = []
            
        # check if remaining is an inline comment and add comment
        post_comment = self._parse_inline_comment(remainder)
        if post_comment:
            self.curent_ttable.add_comment( post_comment, post=True)

    
    def _parse_array_of_tables(self, line):
        """
        Parse array of tables (== [ {}, {} ...]) with elements defined by 
        [[array_of_tables]]
        key = value
        """
        # Split out the full table name from inside []:
        # can be quoted and dotted...
        # e.g. [table."[sub].".table]
        #strip initial '[[' 
        names, remainder = self._get_names(line[2:], term=']]')
        #print(f'names {names}, remainder {repr(remainder)}')
        
        #get the nested tables - but not the last one as this needs to be
        # a list!
        dtable = self._get_data_table(names[:-1], self.data)
        ttable = self._get_template_table(names[:-1], self.template)
        n = names[-1]
        if n not in dtable:
            #create a new list and 1st table/dict inside
            dtable[n] = [{}] 
            ttable[n] = TableArray( [ TableType(),] )
            
            #update current tables
            self.curent_dtable = dtable[n][0]
            self.curent_ttable = ttable[n][0]
            
        elif isinstance(dtable[n], list):
            #add new item to existing list
            dtable[n].append({})
            ttable[n].append( TableType() )
            #update current tables
            self.curent_dtable = dtable[n][-1]
            self.curent_ttable = ttable[n][-1]
            
        else:
            raise ParsingError(f'Name "{n}" is not an array of tables at line number {self.linenum}; {line}')

        #save any precomments
        if self.pre_comments:
            self.curent_ttable.set_comments( self.pre_comments)
            self.pre_comments = []
            
        # check if remaining is an inline comment and add comment
        post_comment = self._parse_inline_comment(remainder)
        if post_comment:
            self.curent_ttable.add_comment( post_comment, post=True)

    # Helper functions working on a line with the inital key = stripped or 
    # inside an array
        
    def _parse_value(self, line):
        """
        Parse a line string containing a value to determine type
        
        Returns (value, ttype, remainder)
        """
        line = line.lstrip()  #strip any leading white space
    
        # ----------------------------------------------------------------------
        # Multi-line types - pass whole line
        # ----------------------------------------------------------------------
        if line.startswith('"""'):
            #multiline string
            return self._parse_ml_string(line)
            
        elif line.startswith("'''"):
            #multiline literal string 
            return self._parse_ml_literal(line)
            
        elif line.startswith("["):
            #array/list
            return self._parse_array(line)
        
        elif line.startswith("{"):
            #in-line table
            return self._parse_inline_table(line)
            
        # ----------------------------------------------------------------------
        # Single line types - but need whole line to test
        # ----------------------------------------------------------------------
        elif line.startswith('"'):
            #string - pass whole line
            return self._parse_string( line)
            
        elif line.startswith("'"):
            #literal string - pass whole line
            return self._parse_literal_string( line)
        
        # ----------------------------------------------------------------------
        #Next test for values on single line after spliting 
        # ----------------------------------------------------------------------
        #print(f'parse_value line {repr(line)}')

        #[] values to split at  ',', '#' '\n' ']' '}'
        #() keep seperator
        try:
            rawvalue, sep, remainder = re.split(r'([,#\n}\]])',line, 1)
        except:
            rawvalue = line
            sep = ''
            remainder =''
        rawvalue = rawvalue.strip()
        remainder = sep+remainder #keep seperator for parsing

        #print(f'parse_value rawvalue {repr(rawvalue)};  remainder {repr(remainder)}')
        
        #a leading comma, ] or \n was found?
        if rawvalue=='':
            return (None, None, remainder)

        # ----------------------------------------------------------------------
        # Single line types - using pre-split value,remainder
        # ----------------------------------------------------------------------
        if rawvalue in ['True', 'true']:
            #bool 
            value = True
            ttype = BoolType()
            
        elif rawvalue in ['False', 'false']:
            #bool
            value = False
            ttype = BoolType()
            
        elif rawvalue.count(':') or rawvalue[1:].count('-')>=2:
            #datetime
            # date: YYYY-MM-DD floats: -1e-5 hence ignore 1st char -
            # time: HH:MM:SS
            value = self._parse_datetime(rawvalue)
            ttype = DateTimeType()
        
        elif rawvalue.startswith( ('0b', '-0b','+0b') ):
            #binary int 0b01
            value = self._parse_int(rawvalue, base=2)
            ttype = IntType(base=2)
            
        elif rawvalue.startswith( ('0o', '-0o','+0o') ):
            #octal int 0o01
            value = self._parse_int(rawvalue, base=8)  
            ttype = IntType(base=8)
            
        elif rawvalue.startswith(('0x', '-0x','+0x') ):
            #hex int 0x01
            value = self._parse_int(rawvalue, base=16)
            ttype = IntType(base=16)
            
        elif rawvalue.count('.') or rawvalue.count('e') or rawvalue.count('E'):
            #float
            # 1.0
            # 1e2
            # 1E2
            value = self._parse_float(rawvalue)
            ttype = FloatType()
        
        elif rawvalue in ['inf', '-inf', '+inf', 'nan', '-nan', '+nan']:
            value = self._parse_float(rawvalue)
            ttype =  FloatType()
            
        else:
            #normal int - try a normal int
            value = self._parse_int(rawvalue, base=10)
            ttype = IntType(base=10)
            
        # ----------------------------------------------------------------------
        # still None raise an exception
        # ----------------------------------------------------------------------
        if value is None:
            raise ParsingError(f'Unparseable value at line number {self.linenum} : {line} ')
        
        return value, ttype, remainder

    
    def _parse_ml_string(self, line):
        """
        Parse multi-line string (triple \") - reading more if required.
        """
        #check for string start
        if line.startswith('"""') is False:
            return None, line
            
        #strip 1st newline (as per toml spec)
        if line[3]=='\n':
            #need to start with next line
            try:
                line = self._read_more()
            except EOFError:
                raise ParsingError(f'Unclosed multiline string at line number {self.linenum}: {line}')
        else:
            #otherwise trim starting quotes
            line = line[3:]

        #find the first sequence of unescaped """
        i=-1
        while True:
            #find next triple quote
            i = line.find('"""', i+1) 
            if i==-1:
                #read next line
                try:
                    newline = self._read_more()
                except EOFError:
                    raise ParsingError(f'Unclosed multiline string at line number {self.linenum}: {line}')
                #strip whitespace when lines end with \
                if line.endswith('\\\n'):
                    line = line.rstrip('\\\n')
                    newline = newline.lstrip()
                line = line+newline
                
            #check if unescaped end found
            elif line[i-1]!='\\':
                break
                
            #print(f'ml_string next {repr(line)}')                

        # split out value and remainder
        value = line[:i]
        remainder = line[i+3:]
        
        #hack check for possible extra quotes upto |""|"""
        if remainder.startswith('""'):
            value = value + '""'
            remainder = remainder[2:]
        elif remainder.startswith('"'):
            value = value + '"'
            remainder = remainder[1:]
        
        try:
            #value = str(value)
            #fix to keep \r inside strings replace with escape sequence
            value = value.replace('\r','\\r')
            value = literal_eval('"""'+value+'"""')
        except Exception as e:
            raise ParsingError( f"{e} at line number {self.linenum} : {line}") 
        
        #print(f'ml_string {repr(value)} , remainder {repr(remainder)}')
        ttype = StringType()
        
        return value, ttype, remainder

    
    def _parse_string(self, line):
        """
        Parse string "" type from string value.
        """
        value, remainder = self._get_string(line)
        #print(f'[{repr(value)}] {value}')
        #remove escaped chars
        try:
            #value = str(value)
            value = literal_eval('"'+value+'"')
        except Exception as e:
            raise ParsingError( f"{e} at line number {self.linenum} : {line}") 
            
        ttype = StringType() ##todo return string ttype
            
        return value, ttype, remainder
       
    
    def _parse_literal_string(self, line):
        """
        Parse literal string '' type from string value.
        """
        #check for literal string
        if line.startswith("'") is False:
            return None, line
            
        #find the first quote close
        i=line.find("'",1)      
        if i==-1:
            raise ParsingError(f"Unclosed literal string \"'\" at line number {self.linenum} : {line}") 
        # split out value and remainder
        remainder = line[i+1:]
        value = line[1:i]
        
        ttype = LiteralStringType()
            
        return value, ttype, remainder
    
    
    def _parse_ml_literal(self, line):
        """
        Parse multi-line literal string (triple ''') - reading more if required.
        """
        #check for string start
        if line.startswith("'''") is False:
            return None, line
            
        #strip 1st newline (as per toml spec)
        if line[3]=='\n':
            #need to start with next line
            try:
                line = self._read_more()
            except EOFError:
                raise ParsingError(f'Unclosed multiline literal string at line number {self.linenum}: {line}')
        else:
            #otherwise trim starting quotes
            line = line[3:]

        #find the first sequence of unescaped """
        i=-1
        while True:
            #find next triple quote
            i = line.find("'''", i+1) 
            if i==-1:
                #read next line
                try:
                    line = line + self._read_more()
                except EOFError:
                    raise ParsingError(f'Unclosed multiline literal string at line number {self.linenum}: {line}')
            #check if unescaped end found
            else:
                break
                
            #print(f'ml_literal next {repr(line)}') 
        
        # split out value and remainder
        value = line[:i]
        remainder = line[i+3:]
        
        #hack check for possible extra quotes upto |''|'''
        if remainder.startswith("''"):
            value = value + "''"
            remainder = remainder[2:]
        elif remainder.startswith("'"):
            value = value + "'"
            remainder = remainder[1:]
            
        #print(f'ml_literal {repr(value)} , remainder {repr(remainder)}')
        ttype = LiteralStringType()
        return value, ttype, remainder
        
    
    def _parse_array(self, line):
        """
        Parse an array of values - as inline arrays can contain comments this
        returns the parse list and a ArrayType list containing any comments.
        """
        #check for array start
        if line.startswith('[') is False:
            return None, line
            
        #itteratively loop over the line
        array = [] #use a list as toml arrays can be mixed 
        ttype = ArrayType() #shadow list with tomltypes
        remainder = line[1:]
        pre_comments = []
        while True:
            
            # strip any white space
            remainder = remainder.lstrip()
            
            # Pre-checks
            # ----------
            # blank line
            if remainder == '':
                #read next line
                try:
                    remainder = self._read_more()
                    remainder = remainder.lstrip()
                except EOFError:
                    raise ParsingError(f'Unclosed array at line number {self.linenum}: {line}')
                    
            # check for comment before value
            if remainder.startswith('#'):
                comment = remainder.strip()
                remainder = ''
                pre_comments.append(comment)
            
            # check for the end of the array!
            elif remainder.startswith(']'):
                remainder = remainder[1:] #remove ]
                break  
                
            # Read a value
            else:
                #print(f'parse_array next loop remainder {repr(remainder)}')
                value, vttype, remainder = self._parse_value(remainder)
                #print(f'parse_array value {repr(value)};  remainder {repr(remainder)}')
                
                # post read value checks
                # ----------------------
                # strip any whitespace before the comma...
                remainder = remainder.lstrip()
            
                # strip comma and whitespace
                if remainder.startswith(','):
                    remainder = remainder[1:].lstrip() #remove , and whitespace
            
                # check for comment after item
                if remainder.startswith('#'):
                    post_comment = remainder.strip()
                    remainder = ''
                else:
                    post_comment = None

                # store comments with the item
                if (vttype is None):
                    vttype = TomlType()
                
                if pre_comments:
                    #print('storing pre-comments', pre_comments)
                    vttype.set_comments(pre_comments)
                    pre_comments = [] #reset for next item
                
                if post_comment:
                    #print('storing post-comments', post_comment)
                    vttype.add_comment( post_comment, post=True)
                
                # finally store the found item/comments
                if value is not None: #value can be None  from parse_value
                    array.append(value)
                    ttype.append(vttype)
            #---
        #---
        #print(f'parse_array return {array};  remainder {repr(remainder)}')
        return array, ttype, remainder        

    
    def _parse_inline_table(self, line):
        """
        Parse inline tables e.g. key = { name= 'tom', age : 7.5 }
        """
        #check for array start
        if line.startswith('{') is False:
            return None, line
        
        #itteratively loop over the line
        table = {}
        ttype = TableType()
        ttype.set_inline(True)
        pre_comments = []

        remainder = line[1:]
        while True:

            # strip any white space
            remainder = remainder.lstrip()
             
            # Pre-checks
            # ----------
            # blank line
            if remainder == '':
                #read next line
                try:
                    remainder = self._read_more()
                    remainder = remainder.lstrip()
                except EOFError:
                    raise ParsingError(f'Unclosed in-line table at line number {self.linenum}: {line}')
                    
            # check for comment before value
            if remainder.startswith('#'):
                comment = remainder.strip()
                remainder = ''
                pre_comments.append(comment)
            
            # check for the end of the table!
            elif remainder.startswith('}'):
                remainder = remainder[1:] #remove ]
                break  
                
            # Read a value
            else:
                #print(f'parse_inline_table next loop; remainder {repr(remainder)}')
                key, value, vttype, remainder = self._parse_key_value(remainder)
                #print(f'parse_inline_table found;key {key}; value {repr(value)};  remainder {repr(remainder)}')

                # post read value checks
                # ----------------------
                # strip any whitespace before the comma...
                remainder = remainder.lstrip()
            
                # strip comma and whitespace
                if remainder.startswith(','):
                    remainder = remainder[1:].lstrip() #remove , and whitespace
            
                # check for comment after item
                if remainder.startswith('#'):
                    post_comment = remainder.strip()
                    remainder = ''
                else:
                    post_comment = None

                # store comments with the item
                if (vttype is None):
                    vttype = TomlType()
                
                if pre_comments:
                    #print('storing pre-comments', pre_comments)
                    vttype.set_comments(pre_comments)
                    pre_comments = [] #reset for next item
                
                if post_comment:
                    #print('storing post-comments', post_comment)
                    vttype.add_comment( post_comment, post=True)
                
                #check key and store
                if key in table:
                    raise ParsingError(f'key {key} already exists in in-line table at linenuymber {self.linenum}: {line}')
                else:
                    table[key] = value
                    ttype[key] = vttype
        
        #print(f'parse_inline_table return {table};  remainder {repr(remainder)}')
        return table, ttype, remainder
       
    
    def _parse_key_value(self, line):
        """
        Parse a key/value pair to toml rules
        """
        # Check for dotted key names that indicate value should be stored in a 
        # table
        # can be quoted and dotted...
        # e.g. 'table."=.=subtable=.=".key = "value" #comment\n'
        #print(f'parse key/value: {line}')
        
        # check for quoted keys 
        if line.startswith( ("'", '"') ):
            key, remainder = self._get_string( line )
            if remainder.lstrip()[0] !='=':
                raise ParsingError(f'Unrecognised line, at line number {self.linenum}; {line}')
            _, raw_value = remainder.split('=', 1)
        else:
            #check for key = value syntax and split
            has_equal = line.count('=')>0
            if not has_equal:
                #not a valid line entry
                raise ParsingError(f'Unrecognised line, at line number {self.linenum}; {line}')
            key, raw_value = line.split('=', 1)
            key = key.strip()           #tidy up the key
            
        #print(f'parse key {key} : {repr(raw_value)}')

        #determine type
        value, ttype, remainder = self._parse_value(raw_value)    
        
        return key, value, ttype, remainder
        
    # Helper functions working on tidied value strings 
    # - comment removed, whitespace stripped
    # - return None if unsucessfull
        
    def _parse_bool(self, value):
        if (value in [ 'true', 'True']):
            v = True
        elif (value in ['false', 'False']):
            v = False
        else:
            v = None
        return v

    
    def _parse_int(self, value, base=10):
        try:
            v = int(value, base)
        except:
            v = None
        return v

    
    def _parse_float(self, value):
        try:
            v = float(value)
        except:
            v = None
        return v
    
    
    def _parse_datetime(self, value):
        try:
            v = datetime_parser(value)
        except:
            v = None
        return v
    
    
    def _parse_inline_comment(self, remainder):
        """check any remainder on a line for a inline comment and add if needed"""
        comment = remainder.strip()
        #print(f'parse inline comment {comment}')
        if comment=='':
            return None
        elif comment.startswith('#') is False:
            raise ParsingError(f'Non-comment content after value at line number {self.linenum}; {remainder}')
        else:
           return comment

#%%-----------------------------------------------------------------------------
# TomlWriter
#-------------------------------------------------------------------------------
class TomlWriter():
    """
    TomlWriter - writes to a toml compatible syntax:

    Should support almost full toml spec now however there may be some issues.
    + bare names can be any valid utf-8 chars (toml only contain ASCII letters,
     ASCII digits, underscores, and dashes (A-Za-z0-9_-)
    
    + This allows table names to have spaces (not in toml spec) don't use them
    if you want to be 100% compatible
    + Inline tables can have newlines inside {} (not in toml spec) but really 
    normal tables should be used.
    """ 
     
    def save(self, filename, data, template=TableType(), indent_subtables=4):
        """
        Write the data dictionary to the filename using the optional
        template given.
        
        Optional kwargs:
        
        indent_subtables - Number of characters to indent subtable by.
        """
        s = template.to_str(  data, indent_subtables=indent_subtables)
        with open( filename, 'w', newline='', encoding='utf-8') as file:
            file.write(s)
            
    
    def dumps(self, data, template=TableType(), indent_subtables=4):
        """
        Write the data dictionary to a string using the optional
        template given.
        
        Optional kwargs:
        
        indent_subtables - Number of characters to indent subtable by.
        """
        #start encoding main dictionary/table
        s = template.to_str(  data, indent_subtables=indent_subtables)

        return s
    
    
    def dump(self, file, data, template= TableType(), indent_subtables=4):
        """
        Write the data dictionary to the open file object using the optional
        template given.
        
        Optional kwargs:
        
        indent_subtables - Number of characters to indent subtable by.
        """
        s = template.to_str(  data, indent_subtables=indent_subtables)
        file.write(s)

#%%-----------------------------------------------------------------------------
# Create interface instance / functions
#-------------------------------------------------------------------------------
toml_parser = TomlParser()
toml_writer = TomlWriter()

def load(filename, return_template=False):
    """ 
    Open the filename give and parses
    
    returns data dictionary and optional template
    """
    data,template = toml_parser.open(filename)
    if return_template:
        return data, template
    else:
        return data

def save(filename, data, template=TableType(), indent_subtables=4):
    """
    Write the data dictionary to the filename
    
    Optional kwargs:
    
    template = type.TableType instance with key templating classes
    indent_subtables - Number of characters to indent subtable by.
    """
    s = template.to_str(  data, indent_subtables=indent_subtables)
    with open( filename, 'w', newline='', encoding='utf-8') as file:
        file.write(s)


"""
TOML types
-----------

Types classes - objects used to store templating information (comments, 
formatting and validation information) and to deal with writing values to string

Also includes the Exceptions and some utility functions used in the to_str 
methods and main encoder class

TODO: 
string formatting options
datetime formatting / validation options
"""
import os
import datetime
import string
import numpy as np  #for handling numpy data types

#%%-----------------------------------------------------------------------------
# Exceptions
#-------------------------------------------------------------------------------
class EncodingError(Exception):
    pass
 
# Define exception to raise when parsing fails.
class ParsingError(Exception):
    pass
 
class ValidationError(Exception):
    pass
    
#%%-----------------------------------------------------------------------------
# utility function
#-------------------------------------------------------------------------------
VALID_TYPES = [bool, int, float, str, datetime.datetime, dict, list, tuple]

def _quote_name( name, allowed = set(string.ascii_letters+string.digits+'_-') ):
    """ Return a quote version of the name if required """
    #check if empty name stirng
    if name=='':
        return '""'
        
    #check if allowed wihtout quotes
    sname = set(name)
    if (sname <= allowed):
        return name
    
    #check if contains quotes
    has_single = "'" in sname
    has_double = '"' in sname
    
    if has_single and not has_double:
        #wrap in double quotes
        return '"'+name+'"'
    if has_double and not has_single:
        #wrap in single quotes
        return "'"+name+"'"
    
    #has both - wrap in double quotes and escape any "
    return '"' + name.replace('"', r'\"')+ '"'
        
def _is_table_array(value):
    """
    check if list value should be treated as an array of tables i.e. all 
    elements are dicts
    """
    if isinstance(value, (list,tuple) ) is False:
        return False
    return all(isinstance(elem ,dict) for elem in value)

def _is_inline_array(value):
    """
    check if list value should be treated as an inline array i.e. some
    elements are not dicts
    """
    if isinstance(value, (list,tuple) ) is False:
        return False
    return not all(isinstance(elem ,dict) for elem in value)
    
def _get_default_type( value ):
    """
    Get a new template object for a value or None
    """
    # Arrays can be inline array or TableArray
    if isinstance( value, (list,tuple)):
        if _is_table_array( value):
            return TableArray()
        else:
            return ArrayType()
            
    #numpy data types need to be converted to python float/int first
    if isinstance(value, np.generic):
        value = value.item()
        
    # Rest are easy
    DEFAULT_TYPES = {   
                    int : IntType,
                    float: FloatType,
                    bool: BoolType,
                    datetime.datetime: DateTimeType,
                    str : StringType,
                    dict: TableType,                 
                    }
    t = DEFAULT_TYPES.get( type(value), None)
    if t is None:
        return None
        
    return t()

def _write_dict(ttable, dtable, 
                    indent=0, indent_subtables=4, table_name=None):
    """
    Utility function to write a dictionary of values - used for Table and 
    TableArray.
    
    ttable  - the template dict
    dtable  - the data dict
    
    indent - current indent in characters
    indent_subtables - Number of characters to indent subtable by
    table_name - None or ['table_name'] to write as
    """
    # ----------------------------------------------------------------------
    # sort keys to next ensure all keys are written before subtables
    # in keeping with toml syntax
    # ----------------------------------------------------------------------
    subs = []
    keys = []
    for key in dtable:
        value = dtable[key]
        
        #sub tables
        if isinstance(value, dict):
            #check if formatted to be written inline
            ttype  = ttable.get( key, TableType() )
            if ttype.inline is False:
                subs.append(key) #write as table
            else:
                keys.append(key) #write as inline table
                
        # check for arrays of tables
        elif isinstance(value, (list,tuple)):          
        
            #check if formated to be written as an array
            ttype  = ttable.get( key, None )
            if (ttype is None) and (_is_table_array(value)):
                ttype = TableArray()
            elif (ttype is None):
                ttype = ArrayType()
            
            if isinstance(ttype, TableArray):
                subs.append(key) #an array of tables
            else:
                keys.append(key) #an inline table

        # everything else
        else:
            keys.append(key)
            
    # ----------------------------------------------------------------------        
    # loop over keys and write first
    # ----------------------------------------------------------------------  
    s = ''
    sindent = ' '*indent      
    for key in keys:

        # get the value and tomltype
        value = dtable[key]
        tvalue = ttable.get( key, None)
        if tvalue is None:
            tvalue = _get_default_type(value)
        if tvalue is None:
            raise EncodingError(f'unknown type in dict {key}{type(value)}')
            
        # check if keyname needs quotes
        key = _quote_name(key)
        
        # write precomments for item
        precomments =  (os.linesep + sindent).join(tvalue.pre_comments) 
        if precomments:
            s +=  sindent + precomments + os.linesep 

        # get string representation - indent passed to multiline types to 
        # enable indenting to start of value in key = value
        svalue = tvalue.to_str(value, indent+len(key)+3)
        
        # write key = value
        s += f'{sindent}{key} = {svalue}'
        
        # write post comment and indent new line
        if tvalue.post_comment is not None:
            s += sindent + tvalue.post_comment + os.linesep
        else:
            s+= os.linesep

    # ----------------------------------------------------------------------
    # loop over subsections (tables, array of tables)
    # ----------------------------------------------------------------------
    # calculate new indent for subtables
    sub_indent = indent + indent_subtables
    for key in subs:        
        # get value and tomltype
        value = dtable[key]
        tvalue = ttable.get(key, None)
        if tvalue is None:
            tvalue = _get_default_type(value)
        if tvalue is None:
            raise EncodingError(f'Unknown type in dict {key}{type(value)}')                

        # get full section name
        if table_name:
            name = table_name+'.'+ _quote_name(key)
        else:
            name =  _quote_name(key)

        # add the subtable string
        s += tvalue.to_str( value,  indent=sub_indent, 
                                    indent_subtables=indent_subtables,
                                    table_name=name)

    return s
#%%-----------------------------------------------------------------------------
# Type classes define how to write, Type and validate items in the
# dictionaries
#-------------------------------------------------------------------------------

#Base Type class only
class TomlBaseType():
    def __init__(self, **kwargs):
        """
        TomlType base type - not used directly
         - implements comments/validation interface
        """
        # Comments
        self.pre_comments = []   #list of comments written before the item
        self.post_comment = None  #list of comments written after the item  
        
    def __repr__(self):
        return str(self.__class__.__name__.rstrip('Type'))

    # Comments interfaces
    def set_comments(self, comments, post=False):
        """
        Set the list of comments assoicated with the item
        post false - written before the item
        post True  - writen after the item (or inline)
        """
        if post:
            self.post_comments = comments
        else:
            self.pre_comments = comments
    
    def add_comment(self, comment, post=False):
        """
        Add a comment.
        post false - written before the item
        post True  - writen inline after the item
        """
        comment = str(comment)
        if comment.startswith('#') is False:
            comment = '# '+comment.lstrip()
        if post:
            self.post_comment=comment
        else:
            self.pre_comments.append(comment)
            
    def get_comments(self, sep=os.linesep):
        """
        Get pre and post comments strings
        """
        pre = os.linesep.join(self.pre_comments) 
        post = self.post_comment+os.linesep
        return pre,post
    
    # Validation interface
    def validate(self, value):
        """
        Check if the value passed meets the templated requirements .
        
        If the value fails type specific tests (i.e. min, max or allowed values)
        this should raise a ValidationError.
        """
        return True
    

# Generic type class used when a better class doesn't exist
class TomlType(TomlBaseType):
    """
    Class representing formating/validation infomation about a keys value
    """
    def __init__(self, **kwargs):
        """
        TomlType base class
        
        Validation kwargs:
        ------------------
        default - default value
        valid = [] list of valid values
        """
        # base init
        TomlBaseType.__init__(self, **kwargs)
        
        # mew attributes
        self.valid = None       #valid value list or None
        self.default = None
        
        # set valid values
        if 'valid' in kwargs:
            self.set_valid( kwargs['valid']) 
        # set default
        if 'default' in kwargs:
            self.set_default( kwargs['default'])
            
    # Formatting interface    
    def to_str(self, value, indent=0):
        """
        Convert the value to a toml style string representation applying any 
        formating options
        """
        return repr(value) 
        
    def from_str(self, svalue):
        """
        Convert from a string representation, i.e. as returned from GUI controls
        """
        return svalue
        
    # Validation interface
    def validate(self, value):
        """
        Check if the value passed meets the templated requirements .
        
        If the value fails type specific tests (i.e. min, max or allowed values)
        this should raise a ValidationError.
        """
        # check value against valid types
        if type(value) not in VALID_TYPES:
            raise ValidationError(f'Unsupported type {type(value)} for TOML Config - should be bool, int, float, str, datetime.datetime')
        
        # check against allowed valid list
        if self.valid is not None:
            if value not in self.valid:
                raise ValidationError(f'Value {value} not in valid list {self.valid}')
        return True
    
    def set_valid(self, valid):
        """ Set a list of valid values or None to clear """
        # clear valid list check
        if valid is None:
            self.valid = None
            return
        
        # cache old valid list incase of failure
        old = self.valid
        self.valid = None
        # check new values against other tests
        for value in valid:
            try:
                self.validate(value)
            except ValidationError as e:
                self.valid = old
                raise ValueError(f'Value {value} in valid list fails validation conditions: {e}')
        
        #store new valid list
        self.valid = valid

    # Default value
    def get_default(self):
        """ Get the default value or None if not set"""
        return self.default
        
    def set_default(self, value):
        """ Set the default value or None to clear """
        if value is None:
            self.default = None
        else:
            self.validate(value)
            self.default = value

# -----------------------------------------------------------------------------
# Toml Types
# -----------------------------------------------------------------------------
#Boolean
class BoolType(TomlType):
    def to_str(self, value, indent=0):
        """
        Convert the value to a toml style string representation applying any 
        formating options
        """
        if value is True:
            value = 'true'
        else:
            value = 'false'
        return f"{value}"

    def from_str(self, svalue):
        """
        Convert from a string representation, i.e. as returned from GUI controls
        """
        if svalue == 'true': value=True
        elif svalue == 'false': value=False
        else: raise ParsingError(f'Cannot convert string {svalue} to bool')
        return value

    def validate(self, value):
        """
        Check if the value passed is okay to interprete as a bool
        """
        if isinstance(value, bool) is False:
            raise ValidationError(f'Expected booleen type, got {type(value)}')
        return True
    
# General numerical type
class NumericalType(TomlType):
    def __init__(self, **kwargs):
        """
        Numerical type can be int/float
        
        Validation kwargs:
        ------------------
        default = set default value
        min = set minimium value
        max = set maximium value
        valid = [] list of valid values
        
        Formating kwargs:
        -----------------
        format = python format string for output e.g. 'd' or 'f'
        """
        # new attributes
        self.min = None
        self.max = None
        self.fstring = '{value:}'          #formating control for output

        # base init
        TomlType.__init__(self, **kwargs)

        # set min/max
        if 'min' in kwargs:
            self.set_min( kwargs['min'])
        if 'max' in kwargs:
            self.set_max( kwargs['max'])
        if 'format' in kwargs:
            self.set_format(format = kwargs['format'])
            
    def from_str(self, svalue):
        """
        Convert from a string representation, i.e. as returned from GUI controls
        - general numerical type allways converts to float
        """
        value = float(svalue)
        return value
        
    def set_min(self, value):
        """Set the minimium valid value or None to clear"""
        if value is None:
            self.min = None
        elif isinstance( value, (int,float)) is False:
            raise ValueError('Expected int/float for limit')
        else:
            self.min = value
    
    def set_max(self, value):
        """Set the maximium valid value or None to clear"""
        if value is None:
            self.max = None
        elif isinstance( value, (int,float)) is False:
            raise ValueError('Expected int/float for limit')
        else:
            self.max = value
    
    def set_format(self, format='0.2f'):
        """
        Set the output format using python string format notation
        This should be valid for float.
        e.g. '0.2f' to_str(255) -> '255.00'
        """
        fstring = '{value:'+str(format)+'}'
        #test
        fstring.format(value = 255.12345)
        #store
        self.fstring = fstring
        
    def validate(self, value):
        """
        Check if the value passed is valid
        """
        # check type
        if isinstance(value, (int,float)) is False:
            raise ValidationError(f'Expected int or float type, got {type(value)}')
        
        # check min/max
        if self.min is not None:
            if value<self.min:
                raise ValidationError(f'Value {value} smaller than minimium condition {self.min}')
        if self.max is not None:
            if value>self.max:
                raise ValidationError(f'Value {value} larger than maximium condition {self.max}')    
        
        # do base class validation
        return TomlType.validate(self, value)

# Int class
class IntType(NumericalType):
    def __init__(self, **kwargs):
        """
        int type
        
        Validation kwargs:
        -----------------
        default = set default value
        valid = [] list of valid values
        min = set minimium value
        max = set maximium value
        
        Formating kwargs:
        -----------------
        format = python format string for output e.g. 'd' 
       
        Or combinations of the following keywords
        (these will be ignored if format keyword argument is given)

        base = 2,8, 10, 16 write values in this format e.g. 'b', 'o', 'd', 'x'
        use_sep = True/False write numbers with _ digit seperators  e.g '_d'
        """
        # base init
        NumericalType.__init__(self, **kwargs)

        # formating kwargs
        # set the format base (2 binary, 8 octal, 10 decimal, 16 hexidecimal)
        if 'base' in kwargs:
            base = kwargs['base']
        else: 
            base = 10
        # use digit seperators
        if 'use_sep' in kwargs:
            use_sep = kwargs['use_sep']
        else:
            use_sep = False
        if 'format' in kwargs:
            format = kwargs['format']
        else:
            format = None
        #set the formatting
        self.set_format( base, use_sep, format)
 
    def set_format(self, base=10, use_sep=False, format=None):
        """
        Set the optional foramting used format the string representation of the 
        int.
        
        base    -  i.e. binary=2, octal=8, decimal=10, hexidecimal=16.
        use_sep - write _ digit seperators
        format  - User provided python style format string that is compatible 
                  with the toml syntax. Only the part after the ':' is required.
                  i.e value=255, 
                  format='+d' gives '+255'
                  format='0=+7d' gives '+000255'
                  
                  Note: Use format will ignore other options and format should
                  be toml compatible! see. toml.io
        """
        #User specified format string
        if format:
            fstring = '{value:'+str(format)+'}'
            #test
            fstring.format(value=255)
            #store
            self.fstring = fstring
            return
        
        if use_sep:
            sep='_'
        else:
            sep=''
        
        if base == 2:
            prefix = '0b'
            typecode='b'
        elif base == 8:
            prefix = '0o'
            typecode='o'
        elif base == 10:
            prefix = ''
            typecode='d'
        elif base == 16:
            prefix = '0x'
            typecode='x'
        else:
            raise ValueError('Unsupport integer base for formatting')
        self.fstring = prefix+'{value:'+sep+typecode+'}'
        
    def to_str(self, value, indent=0):
        """
        Convert the value to a toml style string representation applying any 
        formating options
        """
        try:
            self.fstring.format(value=value)
        except:
            print(value)
            raise
        return self.fstring.format(value=value)
        
    def from_str(self, svalue):
        """
        Convert from a string representation, i.e. as returned from GUI controls
        """
        value = int(float(svalue))
        return value
        
    def validate(self, value):
        """
        Check if the value passed is valid
        """
        if isinstance(value, int) is False:
            raise ValidationError(f'Expected int type, got {type(value)}')
        #do numerical type checks
        return NumericalType.validate(self, value)
        
# Float class
class FloatType(NumericalType):
    
    def __init__(self, **kwargs):
        """
        float type
        
        Validation kwargs:
        -----------------
        default = set default value
        valid = [] list of valid values
        min = set minimium value
        max = set maximium value
        
        Formating kwargs:
        -----------------
        format = python format string for output e.g. 'f' 
        """
        # base init
        NumericalType.__init__(self, **kwargs)

        # formating kwargs
        if 'format' in kwargs:
            self.set_format(format = kwargs['format'])
        else:
            self.set_format('')

    def to_str(self, value, indent=0):
        """
        Convert the value to a toml style string representation applying any 
        formating options
        """
        return self.fstring.format(value=value)
        
    def from_str(self, svalue):
        """
        Convert from a string representation, i.e. as returned from GUI controls
        """
        value = float(svalue)
        return
        
    def validate(self, value):
        """
        Check if the value passed is valid
        """
        if isinstance(value, float) is False:
            raise ValidationError(f'Expected float type, got {type(value)}')
        #do numerical type checks
        return NumericalType.validate(self, value)
        
# String class - split into literal string/ string ???
class StringType(TomlType):
    def __init__(self, **kwargs):
        """
        string type
        
        Validation kwargs:
        -----------------
        default = set default value
        valid = [] list of valid values
        """
        # base init
        TomlType.__init__(self, **kwargs)

    def to_str(self, value, indent=0):
        """
        Convert the value to a toml style string representation applying any 
        formating options
        """
        if value.count('\n')>0:
            term = '"""'
        else:
            term = '"'
        #print('in:',repr(value))
        svalue = value.replace('\\', '\\\\')        #escape backslash
        svalue = svalue.replace('"', '\\"')         #escape "
        #print('out: ',repr(svalue))
        return term+svalue+term
    
    def validate(self, value):
        """
        Check if the value passed is valid
        """
        # check type
        if isinstance(value, str) is False:
            raise ValidationError(f'Expected str type, got {type(value)}')
        
        # do base class validation
        return TomlType.validate(self, value)

# LiteralString  - use to write strings as literal strings - encoder default is 
# to treat python strings as toml strings, however literal string can be 
# specified in a template
class LiteralStringType(TomlType):
    def __init__(self, **kwargs):
        """
        literal string type
        
        Validation kwargs:
        -----------------
        default = set default value
        valid = [] list of valid values
        
        Formating kwargs:
        -----------------
        todo:
        """
        # base init
        TomlType.__init__(self, **kwargs)
        
    def to_str(self, value, indent=0):
        """
        Convert the value to a toml style string representation applying any 
        formating options
        """
        #literal strings can only contain upto 2x''
        if (value.count("'''")>0):
            raise EncodingError('Attempt to encode string with three or more single quotes! as literal string')
        
        #use triple quotes for \n or '
        if (value.count('\n')>0) or (value.count("'")>0):
            term = "'''"
        else:
            term = "'"
        return term+value+term

    def validate(self, value):
        """
        Check if the value passed is valid
        """
        # check type
        if isinstance(value, str) is False:
            raise ValidationError(f'Expected str type, got {type(value)}')
        
        # do base class validation
        return TomlType.validate(self, value)
        
class DateTimeType(TomlType):
    def __init__(self, **kwargs):
        """
        DateTime type
        
        Validation kwargs:
        -----------------
        default = set default value
        valid = [] list of valid values
        before = value must be before specified datetime
        after = value must be after specified datetime
        
        formatting kwargs:
        ------------------
        todo: formatstring approach??? valid toml formats only.
        """
        # base init
        TomlType.__init__(self, **kwargs)
        
    def to_str(self, value, indent=0):
        """
        Convert the value to a toml style string representation applying any 
        formating options
        """
        return value.isoformat()
    
    def validate(self, value):
        """
        Check if the value passed is valid
        """
        # check type
        if isinstance(value, datetime.datetime) is False:
            raise ValidationError(f'Expected str type, got {type(value)}')
        
        # do base class validation (valid values check)
        return TomlType.validate(self, value)
    
# ------------------------------------------------------------------------------
# Container types
# ------------------------------------------------------------------------------

# Tables/Dict class
class TableType(TomlBaseType, dict):
    def __init__(self, data={}, **kwargs):
        """
        Template item representing a dict/table in the TOML spec.
                
        data - dict of TomlTypes
        
        validation kwargs:
        -----------------
        require_templated   - True/False. If True validation will fail if 
                                templated key not present
        allow_untemplated   - True/False. If False validation will fail if an
                                untemplated key is present.

        formatting kwargs:
        ------------------
        inline              - True/False write output as inline table
        write_defaults      - Write default values for this table
        """
        # dict init
        dict.__init__(self, data)
        
        # base init
        TomlBaseType.__init__(self)
        
        # formatting attributes
        self.inline = False       
        self.write_defaults = False

        # set inline
        if 'inline' in kwargs:
            self.set_inline(kwargs['inline'])
            
        # set write_defaults
        if 'write_defaults' in kwargs:
            self.set_write_defaults( kwargs['write_defaults'])
            
        # validation attribtues
        self.require_templated = False
        self.allow_untemplated = True

        # set require templated
        if 'require_templated' in kwargs:
            self.set_require_templated(kwargs['require_templated'])

        # set allow untemplated
        if 'allow_untemplated' in kwargs:
            self.set_allow_untemplated(kwargs['allow_untemplated'])
            
    def __repr__(self):
        return 'Table'+dict.__repr__(self)
    
    # --- overload dict methods to only accept TomlBaseTypes
    def __setitem__(self, key, value):
        if isinstance(value, TomlBaseType) is False:
            raise ValueError('Can only add TomlBaseType to TableType')
        dict.__setitem__(self, key, value)

    def update(self, ttable):
        if isinstance(ttable, TableType) is False:
            raise ValueError('Can only update from TableType')
        dict.update(self, ttable)
        
    # formating interface - used when dict is written inline like a value
    def set_inline(self, inline=True):
        """
        Mark this dictionary as written inline
        """
        self.inline = inline

    def set_write_defaults(self, flag=True):
        """
        Write default values for this table
        """
        self.write_defaults = flag

    def to_str(self, dict_value, 
                    indent=0, indent_subtables=4,
                    table_name=None):
        """ 
        Write to string. If formatting option is set to inline write as inline
        table style otherwise write as full table style
        
        Optional kwargs:
        
        indent - current indent in characters
        indent_subtables - Number of characters to indent subtable by
        table_name - None or ['table_name'] to write as
        """
        # if writing defaults get the default dict then update with values 
        # passed in
        if self.write_defaults:
            dtable = self.get_default()
            dtable.update(dict_value)
        # otherwise dtable is just the value passed in
        else:
            dtable = dict_value
        
        if self.inline:
            s = self._to_str_inline(dtable, indent)
        else:
            s = self._to_str(dtable, indent, indent_subtables, table_name)
        return s
        
    def _to_str_inline(self, dtable, indent=0):
        """Helper method - for writing inline table value"""
        # write output 
        sindent = ' '*indent
        s = '{ '        
        for key in dtable:
            # get the value and tomltype
            value = dtable[key]
            ttype = self.get( key, None)
            if ttype is None:
                ttype = _get_default_type(value)
            if ttype is None:
                raise EncodingError(f'unknown type in dict {key}{type(value)}')
            
            # nested inline dicts are possible but weird! make sure written 
            # inline too
            if isinstance(value, dict):
                ttype.set_inline(True)
                
            # get string representation
            svalue = ttype.to_str(value)
            
            # check if keyname needs quotes
            key = _quote_name(key)
            
            # write precomments for item
            precomments =  (os.linesep + sindent).join(ttype.pre_comments) 
            if precomments:
                s += sindent + precomments + os.linesep               
             
            # write key = value
            s += f'{key} = {svalue}, '
             
            # write psot comments and indent new line for next key = value
            if self.post_comment is not None:
                s += self.post_comment + os.linesep + sindent
        
        # remove trailing comma and add }
        s = s[:-2]+' }'
            
        return s
    
    def _to_str(self, dtable, 
                    indent=0, indent_subtables=4,
                    table_name=None):
        """ Helper method - for writing full table style """
        # write table name
        if table_name is None:
            s = ''          #main table no need for newline before
        else:
            s = os.linesep # start with blank line before table

        # calculate indent for table name 
        sindent = ' '*(indent - indent_subtables)
        
        # write precomments for item
        for precomment in self.pre_comments:
            s += sindent + precomment + os.linesep
        
        if table_name is not None:
            # write section header
            s += sindent + f'[{table_name}]'
            
            # write post comments (should only be post comments in name sections!)
            if self.post_comment is not None:
                s += ' '+self.post_comment + os.linesep
            else:
               s += os.linesep

        # write table contents
        s += _write_dict(self, dtable, indent, indent_subtables, table_name)
        return s
        
    # Validation interface
    def set_require_templated(self, flag=True):
        """
        Ensure validation of dictionary fails if a templated key is missing
        """
        self.require_templated = flag
        
    def set_allow_untemplated(self, flag=True):
        """
        Allow untemplated keys in dictionary to be validated
        """
        self.allow_untemplated = flag
    
    def validate(self, dict_value):
        """
        Check if the dict passed in meets the templated requirements .
        
        This will raise a ValidationError on failure
        """
        # check value against valid types
        if isinstance( dict_value, dict) is False:
            raise ValidationError(f'Expected a dict got {type(dict_value)}')
        
        # check each key...
        if self.require_templated:
            # loop over each key in template and ensure it is present, and the 
            # value validates
            for key in self:
                # check key is present in dict to being validated
                try:
                    value = dict_value[key]
                except:
                    raise ValidationError(f'key {key} is required but not present')
                
                #check value against template
                try:
                    tvalue = self[key]
                    tvalue.validate(value)
                except ValidationError as e:
                    raise ValidationError(f'Value of key {key} fails validation: {e}')
                    
        # check other keys
        for key in dict_value:
            # check if key is a string
            if isinstance(key, str) is False:
                raise ValidationError(f'Non-string key {key} in dictionary')
        
            # check if in template
            if (self.allow_untemplated is False) and (key not in self):
                raise ValidationError(f'Untemplated key {key} found in dictionary (use allow_untemplated=True to allow)')
            
        return True
    
    # Get defaults
    def get_default(self):
        """
        Return a dict constructed from default values in the this template dict
        """
        defaults = {}
        for key in self:
            tvalue = self[key]
            if tvalue is not None:
                default = tvalue.get_default()
            defaults[key] = default
        return defaults

# Arrays/lists in the data
class ArrayType(TomlBaseType, list):
    def __init__(self, data=[], **kwargs):
        """
        Template item representing a Array in the TOML spec.
        
        data = list/tuple of TomlTypes
        
        
        Validation kwargs:
        ------------------
        default - default value
        
        allow_untemplated   - True/False. If False validation will fail if extra
                                untemplated values are present.
        """
        # list init
        list.__init__(self, data)
        
        # base init
        TomlBaseType.__init__(self)
        
        # Validation attributes
        self.allow_untemplated = True
        self.default = None
        
        # set default
        if 'default' in kwargs:
            self.set_default( kwargs['default'])
            
        # set allow untemplated
        if 'allow_untemplated' in kwargs:
            self.set_allow_untemplated(kwargs['allow_untemplated'])
        
    def __repr__(self):
        return 'Array'+list.__repr__(self)
        
    # --- overload list method to only accept TomlBaseType
    def __setitem__(self, n, value):
        if isinstance(value, TomlBaseType) is False:
            raise ValueError('Can only add TomlBaseType to TableArray')
        list.__setitem__(self, n, value)
        
    def append(self, object):
        if isinstance(object, TomlBaseType) is False:
            raise ValueError('Can only add TomlBaseType to TableArray')
        list.append(self, object)
        
    def insert(self, index, object):
        if isinstance(object, TomlBaseType) is False:
            raise ValueError('Can only add TomlBaseType to TableArray')
        list.insert(self, index, object)
        
    def extend(self, iterable):
        for value in iterable:
            self.append(value)
            
    # --- formatting         
    def to_str(self, dlist,  indent=0 ):
        """
        Convert the value to a toml style string representation applying any 
        formating options
        """
        # open list
        s = f'[ '
        sindent = ' '*indent

        # loop over elements in list
        for n in range(len(dlist)):
        
            # get the value and tomltype
            value = dlist[n]
            if n<len(self):
                ttype = self[n]
            else:
                ttype = _get_default_type(value)
            if ttype is None:
                raise EncodingError(f'unknown type in array {type(value)}')
            
            # nested dicts are treated as inline
            if isinstance(value, dict):
                ttype.set_inline(True)
                
            # for a list within a list add a newline first
            elif isinstance( value, (list,tuple) ):
                s += os.linesep + sindent
                
            # write precomments for item
            precomments =  (os.linesep + sindent).join(ttype.pre_comments) 
            if precomments:
                s += sindent + precomments + os.linesep               
             
            # get string representation
            svalue = ttype.to_str(value)
            
            # write value 
            s += f'{svalue}, '
             
            # write comments and indent new line for next value
            if ttype.post_comment is not None:
                s += ttype.post_comment + os.linesep + sindent
        
        #remove trailing comma and add ]
        s = s[:-2]+' ]'
            
        return s
    
    # Validation
    def set_allow_untemplated(self, flag=True):
        """
        Allow untemplated values in list/tuple to be validated
        """
        self.allow_untemplated = flag
        
    def validate(self, dlist):
        """
        Check if the dict passed in meets the templated requirements .
        
        This will raise a ValidationError on failure
        """
        if isinstance( dlist, (list,tuple)) is False:
            raise ValidationError(f'Expected a list or tuple, got {type(list_value)}')
        
        # check templated items
        for n in range(0, len(self)):
            tvalue = self[n]
            if len(dlist)>n:
                value = dlist[n]
            else:
                raise ValidationError(f'Missing templated value {n} in array')
            #check value against template
            try:
                tvalue.validate(value)
            except ValidationError as e:
                raise ValidationError(f'Value {n} fails validation: {e}')
        
        # check other items are valid
        if self.allow_untemplated is False:
            if len(dlist) > len(self):
                raise ValidationError(f'Untemplated values in array len(data)={len(dlist)}, len(template)={len(self)}')
        else:
            for n in range(len(self), len(dlist)):
                value = dlist[n]
                tvalue = _get_default_type(value)
                if tvalue is None:
                    raise ValidationError(f'Invalid value of type {type(value)} in array')
                    
                #check value against template
                try:
                    tvalue.validate(value)
                except ValidationError as e:
                    raise ValidationError(f'Invalid value in array: {e}')
        
    # Default value
    def get_default(self):
        """ Get the default value or None if not set"""
        return self.default
        
    def set_default(self, value):
        """ Set the default value or None to clear """
        if value is None:
            self.default = None
        else:
            self.validate(value)
            self.default = value


# Arrays of tables/list of dicts
class TableArray(TomlBaseType, list):
    def __init__(self, data=[], **kwargs):
        """
        Template item representing a list of dicts/array of tables in the TOML 
        spec.
        
        data = list/tuple of TomlTypes

        TableArrays can be templated by adding individual TableTypes or by 
        using the default table Table.Array.default which will be applied to
        all Tables.
        """
        # list init
        list.__init__(self, data)
        
        # base init
        TomlBaseType.__init__(self)
        
        # attributes
        self.default = None     # default value

        # set default
        if 'default' in kwargs:
            self.set_default( kwargs['default'])
            
    
    def __repr__(self):
        return 'TableArray'+list.__repr__(self)
    
    # --- overload list method to only accept TableType
    def __setitem__(self, n, value):
        if isinstance(value, TableType) is False:
            raise ValueError('Can only add TableType to TableArray')
        list.__setitem__(self, n, value)
        
    def append(self, object):
        if isinstance(object, TableType) is False:
            raise ValueError('Can only add TableType to TableArray')
        list.append(self, object)
        
    def insert(self, index, object):
        if isinstance(object, TableType) is False:
            raise ValueError('Can only add TableType to TableArray')
        list.insert(self, index, object)
        
    def extend(self, iterable):
        for value in iterable:
            self.append(value)
    # ---        
    def to_str(self, dlist, 
                    indent=0, indent_subtables=4, table_name=None):
        """ 
        Write to string. If formatting option is set to inline write as inline
        table style otherwise write as full table style
        
        Optional kwargs:
        
        indent - current indent in characters
        indent_subtables - Number of characters to indent subtable by
        table_name - None or ['table_name'] to write as
        """
        s = ''
        sindent = ' '*indent

        #loop over dicts in the list
        for n in range(0,len(dlist)):
            
            # get the dicts from the lists
            dtable = dlist[n]
            
            # if the template has multiple dicts use them, otherwise use the 
            # first table as template for all, if none defined use an empty 
            # table
            if len(self)>n:
                ttable = self[n]
            elif len(self)==1:
                ttable = self[1]
            else:
                ttable = TableType()
            
            # write precomments
            precomments =  (os.linesep + sindent).join(ttable.pre_comments) 
            if precomments:
                s += sindent + precomments + os.linesep
            
            # write section header
            if table_name:
                s += sindent + f'[[{table_name}]]'
                
            # write post comment and indent new line
            if ttable.post_comment is not None:
                s += sindent + ttable.post_comment + os.linesep
                
            # write the dict items
            s += _write_dict(ttable, dtable, 
                                indent, indent_subtables, table_name)
        return s
    
    # Default value
    def get_default(self):
        """ Get the default Table used for Tables in the array """
        return self.default
        
    def set_default(self, value):
        """ Set the default Table used for Tables in the array """
        if value is None:
            self.default = None
        else:
            self.validate(value)
            self.default = value

    def validate(self, dlist):
        """
        Check if the list passed in meets the templated requirements .
        
        This will raise a ValidationError on failure
        """
        if isinstance( dlist, (list,tuple)) is False:
            raise ValidationError(f'Expected a list or tuple, got {type(list_value)}')
        if _is_table_array(dlist) is False:
            raise ValidationError(f'Expected a list or tuple of dicts')
        
        #loop over dicts in the list
        for n in range(0,len(dlist)):
            
            # get the dicts from the lists
            dtable = dlist[n]
            
            # if the template has multiple dicts use them, otherwise use the 
            # first table as template for all, if none defined use an empty 
            # table
            if len(self)>n:
                ttable = self[n]
            elif len(self)==1:
                ttable = self[1]
            else:
                ttable = TableType()
            
            # validate
            ttable.validate( dtable)



"""
TOML encoder
-----------

TomlEncoder class - writes data dictionary to files/strings using a toml 
compatible syntax.

Templating: 
-----------
The encoder also takes an optional template - a  dictlike object (TableType) 
mirroring the data dictionary containing of TomlType objects defined for the 
types in the TOML syntax. This enables formatting and validation of values and
writing comments assoicated with the value/key.


Actual writing is handled by the TomlTypes.to_str() methods starting with a 
TableType instance representing the toplevel table.
"""
import string
import datetime
import os
import io

from . import types       #types for template/formating

#%%-----------------------------------------------------------------------------
# TomlEncoder
#-------------------------------------------------------------------------------
class TomlEncoder():
    """
    TomLEncoder - writes to a toml compatible syntax:

    Should support almost full toml spec now however there may be some issues.
    + bare names can be any valid utf-8 chars (toml only contain ASCII letters,
     ASCII digits, underscores, and dashes (A-Za-z0-9_-)
    
    + This allows table names to have spaces (not in toml spec) don't use them
    if you want to be 100% compatible
    + Inline tables can have newlines inside {} (not in toml spec) but really 
    normal tables should be used.
    """   
    def save(self, filename, data, template=types.TableType(), indent_subtables=4):
        """
        Write the data dictionary to the filename using the optional
        template given.
        
        Optional kwargs:
        
        indent_subtables - Number of characters to indent subtable by.
        """
        s = template.to_str(  data, indent_subtables=indent_subtables)
        with open( filename, 'w', newline='', encoding='utf-8') as file:
            file.write(s)
    
    def dumps(self, data, template=types.TableType(), indent_subtables=4):
        """
        Write the data dictionary to a string using the optional
        template given.
        
        Optional kwargs:
        
        indent_subtables - Number of characters to indent subtable by.
        """
        #start encoding main dictionary/table
        s = template.to_str(  data, indent_subtables=indent_subtables)

        return s
    
    def dump(self, file, data, template=types.TableType(), indent_subtables=4):
        """
        Write the data dictionary to the open file object using the optional
        template given.
        
        Optional kwargs:
        
        indent_subtables - Number of characters to indent subtable by.
        """
        s = template.to_str(  data, indent_subtables=indent_subtables)
        file.write(s)




"""
Configuration file object

A dict-like subclass to manage data and templates seemlessly.

The TomlDecoder returns two nested dictionaries which can be used directly 
however, this class provides a convience wrapper for both.

It returns subdicts together with their template as a new Config instances for
and automatically applys defualts and validation from the template.

Templating:

c = Config()
c['int'] = types.IntType(valid=[2,8,10,16], default=2) 
c['float'] = types.FloatType(min=0, max=255, default=1.0) 
c['string'] = types.StringType(default='test')

sub = types.DictType()
c['section'] = sub
sub['test'] = types.IntType(default=1) 
sub['pi'] = types.FloatType(default=3.14) 

"""
from . import types
from .types import ValidationError
from .parser import toml_parser

#%%-----------------------------------------------------------------------------
# Main config class
#-------------------------------------------------------------------------------
class Config():
    def __init__(self, file=None, data={}, template=None ):
        """
        Config object stores both data/template together with convience   
        
        file - path of toml file to open
        
        or
        
        data - dictionary of values to store in Config
        template - option templating dictionary/TableType
        """
        if file is not None:
            self.load(file)
        else:
            self.data = {}
            if template is None:
                template = {}
            self.template = types.TableType(template)
            self.update(data)
    
    # --- interface methods
    def load(self, file, update=False):
        """
        Load the filename or open file object given.
        
        [optional] update=True to update the current configuration with the 
        loaded values.
        """
        if isinstance(file, str):
            d,t = toml_parser.open(file)
        else:
            d,t = toml_parser.load(file)
        if update:
            self.data.update(d)
            self.template.update(t)
        else:
            self.data = d
            self.template = t
            
    def loads(self, s):
        """
        Parse the string given into this Config object.
        
        [optional] update=True to update the current configuration with the 
        loaded values.
        """
        d,t = TomlParser.loads(s)
        if update:
            self.data.update(d)
            self.template.update(t)
        else:
            self.data = d
            self.template = t
        
    def save(self, filename, indent_subtables=4):
        """
        Save the Config to the filename given.
        """
        s = self.template.to_str(  self.data, indent_subtables=indent_subtables)
        with open( filename, 'w', newline='', encoding='utf-8') as file:
            file.write(s)
        
    def dump(self, file, indent_subtables=4):
        """
        Write the Config to open file object given.
        
        Optional kwargs:
        
        indent_subtables - Number of characters to indent subtable by.
        """
        s = self.template.to_str(  self.data, indent_subtables=indent_subtables)
        file.write(s)   
         
    def dumps(self, indent_subtables=4 ):
        """
        Write the Config to a string.
        
        Optional kwargs:
        
        indent_subtables - Number of characters to indent subtable by.
        """
        s = self.template.to_str(  self.data, indent_subtables=indent_subtables)
        return s
    
    def as_dict(self):
        """
        Return the configuration dictionary only
        """
        return self.data
        
    # --- Basic dict-like interfaces to alow subdicts to retreive together with
    # their template as a new Config instance sharing the same dictionaries
    def __getitem__(self, key):
        """
        Get item:
            - allow only string keys
            - allow multilevel indexing
            - return sub-dicts as Configs
            - return default value if key is templated but not set
            todo:
            - return lists/tuples as ConfigList with template info
        """
        if isinstance(key, str) is False:
            raise KeyError(f'Config keys much be strings {key}')
            
        # get dicts for a multilevel key
        *parents, key = key.split('.')
        table = self.data       # start with data dict
        ttable = self.template  # start with template dict
        for part in parents:
            # get next table
            table = table[part]
            
            # get next template
            ttable = ttable.get(part, {} )

        #get the value and template type
        value = table.get( key, None)
        tvalue = ttable.get( key, None)
        
        # no value - try template for default
        if (value is None) and (tvalue is not None):
            value = tvalue.get_default()
        
        # still no value assoicated with the key
        if value is None:
            raise KeyError(f'{key}')
        
        # wrap as config object if a dict
        if isinstance(value, dict):
            value = Config(data=value, template=tvalue)

        return value
        
    def __setitem__(self, key, value):
        """
        set item:
            - allow multilevel indexing
            - validate key value against template
            - if config.type.TomlType instance store as template
        """
        if isinstance(key, str) is False:
            raise KeyError('Config keys much be strings')
        
        # get dicts for a multilevel key
        *parents, key = key.split('.')
        table = self.data       # start with data dict
        ttable = self.template  # start with template dict
        for part in parents:
            # get next table
            table = table[part]
            
            # get next template (create if needed)
            ttable = ttable.get(part, None )
            if ttable is None:
                ttable = types.TableType()
                ttable[ part ] =  ttable
        
        # Check if setting a Config object
        if isinstance(value, Config):
            table[key] = value.data
            ttable[key] = value.template
            
        # Check if setting the template value
        elif isinstance(value, types.TomlBaseType):
            #p assed value is actually a template value
            tvalue = value
            
            # check if existing value passes validation
            value = table.get( key, None)
            if (value is not None):
                try:
                    tvalue.validate(value)
                except ValidationError as e:
                    raise ValidationError(f'Existing value does not meet new template requirements! {e}')
            # all okay - store new template
            ttable[key] = tvalue
        
        # otherwise setting the value
        else:
            # get the template type
            tvalue = ttable.get( key, None)
            
            # check value against template type
            if tvalue is not None:
                # this will raise ValidationError if value is not acceptable
                tvalue.validate(value)
                
            elif types._get_default_type(value) is None:
                raise ValueError(f'Unsupported type {type(value)} for TOML Config - should be bool, int, float, str, list, tuple, dict, datetime.datetime')
        
            # store
            table[key] = value 

    def __iter__(self):
        return iter(self.data)
        
    def __len__(self):
        return self.data.__len__()
        
    def keys(self):
        return self.data.keys()
    
    def items(self):
        return self.data.items()
        
    def pop(self, key):
        """
        Remove specified key from data and template returning both
        """
        value = self.data.pop(key)
        tvalue = self.template.pop(key)
        return (value, tvalue)

    def update(self, data):
        """
        Update the internal values from the data dictionary given.
        
        This will validate each value against the internal template.
        """
        for key in data:
            #get value
            value = data[key]
            
            # get the template value
            tvalue = self.template.get( key, None)
            
            # check value against template type
            if tvalue is not None:
                # this will raise ValidationError if value is not acceptable
                tvalue.validate(value)
                
            elif types._get_default_type(value) is None:
                raise ValueError(f'Unsupported type {type(value)} for TOML Config - should be bool, int, float, str, list, tuple, dict, datetime.datetime')
        
            # store
            self.data[key] = value 
        
    def get(self, key, default=None):
        """"
        Dictlike get method that will store the default in the dictionary when 
        requested
        """
        try:
            value = self[key]
        except KeyError:
            self[key] = value = default
        return value
        
    # --- pretty printing functions
    def __repr__(self):
        s= 'Config section:\n'
        for key,value in self.items():
            if isinstance(value, dict):
                s +=f'[{repr(key)}]\n'
            #elif isinstance(value, (list, tuple)):
            #    s +=f'{repr(key)} = {repr\n'
            else:
                s +=f'{repr(key)} = {repr(value)}\n'
        return s
        
    # --- extra config interfaces using template
    def add_template(self, key, type, **kwargs):
        pass
        
    def add_comment(self, key, comment):
        pass
        
    def set_from_string(self, key, svalue):
        pass



#!/usr/bin/env python
"""
Python distutils setup.py script:

Usage:
------

 * Recommended: Install via pip / pypi for the latest version *
 
    $ pip install toml_file
 
 
Install as a python package [not manual uninstall will be required]
    $ python setup.py install

Make a wheel and isntall [pip install wheel]
    $ python setup.py sdist bdist_wheel
    $ pip install wheel_file.whl
    
Upload to pypi [requires authorisation]
    $ setup.py upload

editable install
pip install -e /path/to/package
"""
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

#-------------------------------------------------------------------------------
# Add useful commands
#-------------------------------------------------------------------------------
class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds")
            here = os.path.abspath(os.path.dirname(__file__))
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution")
        os.system('"{}" setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status("Uploading the package to PyPI via Twine")
        os.system("twine upload dist//*")

        sys.exit()
        
#-------------------------------------------------------------------------------
# Collect data for the distutils setup() call.
#-------------------------------------------------------------------------------

# Package meta-data.
NAME = "toml_file"
DESCRIPTION = "TOML file reader - python module to read/write toml configuration files"
URL = "https://gitlab.com/tcharrett/totoml"
EMAIL = "tcharrett@gmail.com"
AUTHOR = "Tom Charrett"
REQUIRES_PYTHON = ">=3.7.6"
VERSION = "1.0.0"

# What packages are required for this module to be executed?
REQUIRED = []

# What packages are optional?
EXTRAS = {}

LONG_DESCRIP = """
TOML file reader - python module to read/write toml configuration files

- Files can be loaded and saved completed with comments and formating
- Configuration files can be templated and values validated on load or value changes

Two interaces:

Simple dictionaries {key: value} 
    - data dictionary with nested dictionary with {key: value}
    - template dictionary with comments/type information
    - template allows type, min/max  and other validators to be set and stores 
      comments
    - use load/save functions to open/save to file.

Config object interface:
    - allows seamless use of template/data via custom dictlike interface
    - use cfg = Config(filename) to open
    - access subtables/keys via exteneded key indexing e.g. cfg['Table1.subtable.key'] = 1
    - if a template has been set any value will be validated when it is set.
    - if a template has been set any missing values will return the defualt and be created.
    - use cfg.save(filename) to save
"""

#Find packages to install
PACKAGES = find_packages()

#-------------------------------------------------------------------------------
#Do the setup call
#-------------------------------------------------------------------------------
setup(
    name = NAME,
    version = VERSION,
    description = DESCRIPTION,
    long_description = LONG_DESCRIP,
    author = AUTHOR,
    author_email = EMAIL,
    url = URL,
    packages = PACKAGES,
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Topic :: Scientific/Engineering",
        "Topic :: Text Processing",],
    #upload command
    cmdclass={"upload": UploadCommand},
)
